/* Parameter Sweeping Program, Written 16 June 2011 to 23 June 2011
 * Whitney Sanders (sanderswe@email.wofford.edu)

 * DESCRIPTION:
 * To accompany Age- and Stage-Structured Models, an educational module
 * written by Dr. Angela Shiflet and Dr. George Shiflet.
 * This program executes the given Leslie matrix model based on multiple
 * elements in a set of parameters.

 * INPUT:
 * The input is described in "LeslieInput.dat"
 * The first line contains the length/width of the given Leslie matrix: n
 * The next line contains the Leslie matrix in one-dimensional format
 * (i.e. the first row comes first, then the second row, etc)
 * The next line contains the number of parameters, p
 * The next p lines are the the following format:
 *	 row, column, start, stop, step-size
 * where row and column give the position of the parameter to be changed.

 * OUTPUT:
 * Output is stored in "LeslieOutput.txt"
 * There is one line for each possible combination of values for the
 * p parameters that are changing. Each line is in the following format:
 *	 Value of 1st parameter, ... , Value of pth parameter: lambda
*/

#include <stdio.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>

double getDominantEigenvalue(double* matrix, int n) {
	// Use the GNU Scientific Library (GSL) to calculate the
	// eigenvalues of the given matrix.
	gsl_matrix_view m = gsl_matrix_view_array (matrix, n, n);
	gsl_vector_complex * eval = gsl_vector_complex_alloc(n);
	gsl_matrix_complex * evec = gsl_matrix_complex_alloc(n, n);
	gsl_eigen_nonsymmv_workspace * w = gsl_eigen_nonsymmv_alloc(n);
	gsl_eigen_nonsymmv(&m.matrix, eval, evec, w);
	gsl_eigen_nonsymmv_free(w);

	double dominant=0; // Will contain the dominant eigenvalue
	for (int i = 0; i < n; i++) {
		gsl_complex eval_i = gsl_vector_complex_get (eval, i);
		int isReal = (GSL_IMAG(eval_i)==0);
		double realValue = GSL_REAL(eval_i);
		if (isReal && (fabs(realValue) > dominant) )
			dominant = fabs(realValue);
	}

	// Calculate and return dominant eigenvalue
	return dominant;
}

int main(void) {
	FILE * inFile = fopen("LeslieInput.dat","r");
	FILE * outFile = fopen("LeslieOutput.txt","w");

	int size, p; // Length/width of the matrix, number of parameters
	fscanf(inFile, "%i", &size);

	// Read base Leslie matrix using 1-dimensional array representation
	double leslieMatrix[size * size];
	for (int i = 0; i < size * size; i++)
		fscanf(inFile, "%lf", leslieMatrix+i);

	fscanf(inFile, "%i", &p); // Read number of parameters

	// Put the parameter specifications in arrays
	int positions[p][2]; // Position of each parameter: row, column
	double ranges[p][3]; // For each parameter: start, stop, step

	for (int i = 0; i < p; i++)
		fscanf(inFile, "%i %i %lf %lf %lf", positions[i], positions[i] + 1, ranges[i], ranges[i] + 1, ranges[i] + 2);

	// Note: Float to int conversion uses truncating, while float to
	// long conversion uses rounding.
	// Thus, several of the following variables are declared as longs,
	// since the calculation for the number of steps must use floats,
	// but the number of steps for each parameter is discrete.

	long int steps[p]; // Number of steps for each paramter
	long int total = 1; // Number of matrices that will be constructed
	for (int i = 0; i < p; i++) {
		// Range is inclusive, so we add 1 to step count
		steps[i] = (ranges[i][1] - ranges[i][0]) / ranges[i][2] + 1;
		total = total * steps[i];
	}

	// Generate array of combinations of parameters
	double currentParameterValues[p];
	for (int i = 0; i < p; i++) // Initialize steps at starting position
		currentParameterValues[i] = ranges[i][0];
	double combinations[total][p];
	for (int k = 0; k < p; k++) {
		// Copy the initial parameter values into combinations[0]
		combinations[0][k] = currentParameterValues[k];
	}
	for (int i = 1; i < total; i++) {
		int j = p-1;
		while ( (j > 0) && currentParameterValues[j] == ranges[j][1]) {
			// If a parameter has already reached its stop val,
			// reset it at the start val and move left.
			currentParameterValues[j] = ranges[j][0];
			j -= 1;
		}
		currentParameterValues[j] += ranges[j][2]; // Step once
		for (int k = 0; k < p; k++) {
			// Copy these parameter values into combinations[i]
			combinations[i][k] = currentParameterValues[k];
		}
	} // Completed generating array of combinations of parameters

	// Print first line of output file
	for (int i = 0; i < p - 1; i++)
		fprintf(outFile, "Value at (%i, %i), ", positions[i][0], positions[i][1]);
	fprintf(outFile, "Value at (%i, %i)\n", positions[p - 1][0], positions[p - 1][1]);

	// Calculate new eigenvalue for each combination of parameters
	for (int i = 0; i < total; i++) {
		// Copy the original Leslie matrix
		double matrix[size * size];
		for (int j = 0; j < size * size; j++)
			matrix[j] = leslieMatrix[j];

		// Insert the appropriate parameter values into matrix
		for (int j = 0; j < p; j++) {
			double value = combinations[i][j];
			int row = positions[j][0];
			int column = positions[j][1];
			// Convert to position in the 1-dimensional array
			// representation: we subtract 1 because the input
			// counts rows/columns starting at 1, not 0
			int position = (row - 1) * size + (column - 1);
			matrix[position] = value;
		}

		double dominant_eig = getDominantEigenvalue(matrix, size);

		// Print eigenvalue to output file
		for (int j = 0; j < p - 1; j++)
			fprintf(outFile, "%lf, ", combinations[i][j]);
		fprintf(outFile, "%lf : lambda = %lf\n", combinations[i][p - 1], dominant_eig);
	}

	fclose(inFile);
	fclose(outFile);
	return 0;
}
